
/*MA****************************************************************/
/*                                                                 */
/*     File: symbol-table.h                                        */
/*   Author: Helmut Schmid                                         */
/*  Purpose:                                                       */
/*  Created: Thu Jan  2 14:55:10 2003                              */
/* Modified: Fri Mar  6 12:30:45 2009 (schmid)                     */
/*                                                                 */
/*ME****************************************************************/

#include <string.h>

#include <ext/hash_map>
using __gnu_cxx::hash_map;
using __gnu_cxx::hash;

struct eqstr {
  bool operator()(const char* s1, const char* s2) const {
    return strcmp(s1, s2) == 0;
  }
};


/*****************  class SymbolClass  ***************************/

class SymbolClass {

private:

  typedef hash_map<const char*, int, hash<const char*>, eqstr> SM;

  SM sm;

public:

  int &operator[]( const char *s ) {
    SM::iterator it=sm.find(s);
    if (it != sm.end())
      return it->second;
    return sm.insert(SM::value_type(strdup(s), 0)).first->second;
  }
};


/*****************  class SymbolMap  *****************************/

class SymbolMap {

private:

  typedef hash_map<const char*, const char*, hash<const char*>, eqstr> SM;

  SM sm;

public:

  typedef SM::iterator iterator;

  iterator find( const char *s ) { return sm.find(s); }
  iterator begin() { return sm.begin(); }
  iterator end()   { return sm.end(); }
  size_t   size()  { return sm.size(); }

  const char* &operator[]( const char *s ) {
    iterator it=sm.find(s);
    if (it != sm.end())
      return it->second;
    return sm.insert(SM::value_type(strdup(s), NULL)).first->second;
  }
};


/*****************  class SymbolTable  *****************************/

class SymbolTable {

private:

  typedef hash_map<const char*, size_t, hash<const char*>, eqstr> SymbolMap;

  SymbolMap ST;
  vector<const char*> SN;

public:
  size_t number( const char *s ) {
    SymbolMap::iterator it=ST.find(s);
    if (it != ST.end())
      return it->second;

    // insert the new symbol into the tables
    char *r = strdup(s);
    size_t n = SN.size();
    SN.push_back(r);
    ST[r] = n;
    return n;
  }

  const char *name( size_t n ) {
    if (n < SN.size())
      return SN[n];
    return NULL;
  }

  typedef SymbolMap::iterator iterator;

  iterator find( const char *s ) { return ST.find(s); }
  iterator begin() { return ST.begin(); }
  iterator end()   { return ST.end(); }
  size_t   size()  { return ST.size(); }
};

