
/*MA****************************************************************/
/*                                                                 */
/*     File: tree.C                                                */
/*   Author: Helmut Schmid                                         */
/*  Purpose:                                                       */
/*  Created: Tue Jan 21 13:27:26 2003                              */
/* Modified: Mon Feb 21 11:55:03 2011 (schmid)                     */
/*                                                                 */
/*ME****************************************************************/

#include <stdio.h>
#include <string.h>

#include <iostream>
using std::cerr;


#include "tree.h"
#include "frequency-table.h"

bool Verbose=true;
bool PrintRules=false;
bool WithHeads=false;
bool PrettyPrint=false;
bool DeleteChains=false;
bool PrintTraces=false;
bool UndoMarkov=false;
bool UnaryBranches=false;
bool EmptyCategories=false;
bool ConvertTraces=false;
bool ReorderWords=false;
bool Governors=false;
bool RestoreGF=false;

char *MapFileName=NULL;
char *PrepFileName=NULL;
SymbolMap CatMap;
SymbolMap PrepMap;

FILE *outfile=NULL, *lexfile=NULL;

SymbolClass FeatureClass;

FrequencyTable SubcatFreq;


/*FA****************************************************************/
/*                                                                 */
/*  is_new_subcat                                                  */
/*                                                                 */
/*FE****************************************************************/

bool is_new_subcat( char *pos, char *verb, char *subcat )

{
  char buffer[1000];
  sprintf(buffer,"%s %s %s", pos, verb, subcat);
  return (++SubcatFreq[buffer] == 1);
}


/*FA****************************************************************/
/*                                                                 */
/*  init                                                           */
/*                                                                 */
/*FE****************************************************************/

void init()

{
  FeatureClass["Sg"] = 1;
  FeatureClass["Pl"] = 1;
  FeatureClass["Nom"] = 1;
  FeatureClass["Gen"] = 1;
  FeatureClass["Dat"] = 1;
  FeatureClass["Acc"] = 1;
  FeatureClass["Pos"] = 1;
  FeatureClass["Comp"] = 1;
  FeatureClass["Sup"] = 1;
  FeatureClass["Masc"] = 1;
  FeatureClass["Fem"] = 1;
  FeatureClass["Neut"] = 1;
}


/*FA****************************************************************/
/*                                                                 */
/*  strpref                                                        */
/*                                                                 */
/*FE****************************************************************/

bool strpref( const char *s, const char *p )

{
  return (strncmp( s, p, strlen(p) ) == 0);
}


/*FA****************************************************************/
/*                                                                 */
/*  is_ordinal                                                     */
/*                                                                 */
/*FE****************************************************************/

bool is_ordinal( char *l )

{
  char *s=l;
  while (*s >= '0' && *s <= '9')
    s++;
  return (s > l && *s == '.' && *(s+1) == 0);
}


/*FA****************************************************************/
/*                                                                 */
/*  is_year_number                                                 */
/*                                                                 */
/*FE****************************************************************/

bool is_year_number( char *l )

{
  return (((l[0] == '1' && l[1] == '9' && l[2] >= '0' && l[2] <= '9') ||
	   (l[0] == '2' && l[1] == '0' && l[2] >= '0' && l[2] <= '1')) &&
	  l[3] >= '0' && l[3] <= '9' && l[4] == 0);
}


/*FA****************************************************************/
/*                                                                 */
/*  is_month_name                                                  */
/*                                                                 */
/*FE****************************************************************/

bool is_month_name( char *l )

{
  return (strcmp(l,"January") == 0 || strcmp(l,"February") == 0 ||
	  strcmp(l,"March") == 0 || strcmp(l,"April") == 0 ||
	  strcmp(l,"May") == 0 || strcmp(l,"June") == 0 ||
	  strcmp(l,"July") == 0 || strcmp(l,"August") == 0 ||
	  strcmp(l,"September") == 0 || strcmp(l,"October") == 0 ||
	  strcmp(l,"November") == 0 || strcmp(l,"December") == 0);
}


/*FA****************************************************************/
/*                                                                 */
/*  is_day_of_week                                                 */
/*                                                                 */
/*FE****************************************************************/

bool is_day_of_week( char *l )

{
  return (strcmp(l,"Monday") == 0 || strcmp(l,"Tuesday") == 0 ||
	  strcmp(l,"Wednesday") == 0 || strcmp(l,"Thursday") == 0 ||
	  strcmp(l,"Friday") == 0 || strcmp(l,"Saturday") == 0 ||
	  strcmp(l,"Sunday") == 0);
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::assign_heads                                             */
/*                                                                 */
/*FE****************************************************************/

void Tree::assign_heads()

{
  if (daughters.size() == 1 && daughters[0]->daughters.size() == 0) {
    head_daughter = NULL;
    if (daughters[0]->matching_category("*T*"))
      head = NULL;
    else
      head = this;
    return;
  }

  for( size_t d=0; d<daughters.size(); d++ )
    daughters[d]->assign_heads();

  // general default
  head_daughter = daughters[0];
  head = daughters[0]->head;

  for( size_t d=0; d<daughters.size(); d++ ) {
    if (daughters[d]->matching_base_cat("VP-OC") &&
	(matching_base_cat("S") || matching_base_cat("VP")))
      {
	head_daughter = daughters[d];
	head = daughters[d]->head;
	break;
      }
    else if (matching_base_cat("PP") &&
	     (daughters[d]->matching_base_cat("APPR") ||
	      daughters[d]->matching_base_cat("APPO") ||
	      daughters[d]->matching_base_cat("APPRART")))
      {
	head2 = daughters[d]->head;
      }
    else {
      vector<char*> &s=split_name(daughters[d]->label, 3);
      if (strcmp(s[1],"-HD") == 0 || strcmp(s[1],"-NK") == 0) {
	head_daughter = daughters[d];
	head = daughters[d]->head;
      }
    }
  }
  if (head2) {
    Tree *d = head;
    head = head2;
    head2 = d;
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::strip_features                                           */
/*                                                                 */
/*FE****************************************************************/

void Tree::strip_features()

{
  if (daughters.size() == 0)
    return;
  for( char *p=label; *p; p++ )
    if (*p == '/') {
      *p = 0;
      break;
    }
}


/*FA****************************************************************/
/*                                                                 */
/*  last_char                                                      */
/*                                                                 */
/*FE****************************************************************/

char *last_char( char *s ) {
  if (*s) {
    do { s++; } while (*s);
    s--;
  }
  return s;
}

const char *last_char( const char *s ) {
  if (*s) {
    do { s++; } while (*s);
    s--;
  }
  return s;
}

bool is_digit( char c ) { return c >= '0' && c <= '9'; }


/*FA****************************************************************/
/*                                                                 */
/*  read_mapping                                                   */
/*                                                                 */
/*  Purpose: reads a mapping from Penn tags to (simpler) Penn tags */
/*           from a file                                           */
/*                                                                 */
/*FE****************************************************************/

void read_mapping( char *filename, SymbolMap &map )

{
  FILE *file;
  if ((file = fopen(filename,"rt")) == NULL) {
    cerr << "Error: unable to open map file \"" << filename << "\"!\n";
    exit(1);
  }
  char buffer[1000];
  while (fgets(buffer,1000,file)) {
    char *s1=strtok(buffer," \t\r");
    char *s2=strtok(NULL," \t\r\n");
    if (s2)
      map[s1] = strdup(s2);
  }
  fclose(file);
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::predecessor                                              */
/*                                                                 */
/*FE****************************************************************/

Tree *Tree::predecessor()

{
  if (mother)
    for( size_t i=1; i<mother->daughters.size(); i++ )
      if (mother->daughters[i] == this)
	return mother->daughters[i-1];
  
  return NULL;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::successor                                                */
/*                                                                 */
/*FE****************************************************************/

Tree *Tree::successor()

{
  if (mother)
    for( size_t i=0; i<mother->daughters.size()-1; i++ )
      if (mother->daughters[i] == this)
	return mother->daughters[i+1];
  
  return NULL;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::is_first_daughter                                        */
/*                                                                 */
/*FE****************************************************************/

bool Tree::is_first_daughter()

{
  return (mother && this == mother->daughters[0]);
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::is_last_daughter                                         */
/*                                                                 */
/*FE****************************************************************/

bool Tree::is_last_daughter()

{
  return (mother && this == mother->last_daughter());
}


/*FA****************************************************************/
/*                                                                 */
/*  index_start                                                    */
/*                                                                 */
/*FE****************************************************************/

bool index_start( const char *s )

{
  if (*s != '-' || !is_digit(*(s+1)))
    return false;
  const char *p;
  for( p=s+2; is_digit(*p); p++) ;
  if (*p == 0 || *p == '-')
    return true;
  else
    return false;
}


/*FA****************************************************************/
/*                                                                 */
/*  feature_boundary                                               */
/*                                                                 */
/*FE****************************************************************/

static bool feature_boundary( const char *p )

{
  return (*p == '_' || *p == '+' || *p == '=' || *p == '/' || *p == '.' || 
	  (*p == '%' && *(p-1) != '/') || index_start(p));
}


/*FA****************************************************************/
/*                                                                 */
/*  slash_feature_boundary                                         */
/*                                                                 */
/*FE****************************************************************/

static bool slash_feature_boundary( const char *p )

{
  return (*p == '_' || *p == '+' || *p == '=' ||
	  (*p == '%' && *(p-1) != '/') || index_start(p));
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::matching_category                                        */
/*                                                                 */
/*FE****************************************************************/

bool Tree::matching_category( const char *c )

{
  for( char *p=label; *p; p++ )
    if (*c == 0)
      return feature_boundary(p);
    else if (*p == *c || *c == '?')
      c++;
    else
      return false;
  return *c == 0;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::matching_base_cat                                        */
/*                                                                 */
/*FE****************************************************************/

bool Tree::matching_base_cat( const char *c )

{
  for( char *p=label; *p; p++, c++ ) {
    if (*c == 0)
      return (*p == '-' || feature_boundary(p));
    if (*p != *c && *c != '?')
      return false;
  }
  return *c == 0;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::matching_suffix                                          */
/*                                                                 */
/*FE****************************************************************/

bool Tree::matching_suffix( const char *suffix )

{
  char *l;
  for( l=label; *l; l++ ) ;
  const char *s;
  for( s=suffix; *s; s++ ) ;

  for( ; s>=suffix; s--, l--)
    if (l < label || *s != *l)
      return false;
  return true;
}


/*FA****************************************************************/
/*                                                                 */
/*  append_string                                                  */
/*                                                                 */
/*FE****************************************************************/

static void append_string( char *&p, const char *q )

{
  while ((*(p++) = *(q++))) ;
  p--;
}


/*FA****************************************************************/
/*                                                                 */
/*  lowercase                                                      */
/*                                                                 */
/*FE****************************************************************/

static char *lowercase( const char *word )

{
  static char buffer[1000];
  for( int i=0; i<1000; i++) {
    if ((word[i] >= 'A' && word[i] <= 'Z') || 
	(word[i] >= '' && word[i] <= ''))
      buffer[i] = word[i] + 32;
    else
      buffer[i] = word[i];
    if (word[i] == 0)
      return buffer;
  }
  throw "Error: word too long!\n";
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::change_label                                             */
/*                                                                 */
/*FE****************************************************************/

void Tree::change_label( const char *s )

{
  char *tmp = label;
  label = strdup(s);
  free(tmp);
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::change_base_label                                        */
/*                                                                 */
/*FE****************************************************************/

void Tree::change_base_label( const char *l )

{
  vector<char*> &s=split_name(label, 2);
  char buffer[1000];
  sprintf(buffer,"%s%s", l, s[1]);

  char *tmp = label;
  label = strdup(buffer);
  free(tmp);
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::change_gf_label                                          */
/*                                                                 */
/*FE****************************************************************/

void Tree::change_gf_label( const char *l )

{
  vector<char*> &s=split_name(label, 3);
  char buffer[1000];
  sprintf(buffer,"%s%s%s", s[0], l, s[2]);

  char *tmp = label;
  label = strdup(buffer);
  free(tmp);
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::base_cat                                                 */
/*                                                                 */
/*FE****************************************************************/

const char *Tree::base_cat()

{
  if (matching_base_cat("NP") || matching_base_cat("PP"))
    return "N";
  if (matching_base_cat("VP") || matching_base_cat("S"))
    return "V";
  return "0";
}


/*FA****************************************************************/
/*                                                                 */
/*  find_index                                                     */
/*                                                                 */
/*FE****************************************************************/

char *find_index( char *label )

{
  for( char *p=label; *p; p++ )
    if (index_start(p)) {
      char *q;
      for( q=p+2; *q; q++ )
	if (!is_digit(*q))
	  break;
      if (*q == 0)
	return p;
    }

  return NULL;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::is_trace                                                 */
/*                                                                 */
/*FE****************************************************************/

bool Tree::is_trace()

{
  if (daughters.size() == 0 && mother)
    return matching_category("*T*");

  for( size_t i=0; i<daughters.size(); i++ )
    if (!daughters[i]->is_trace())
      return false;

  return true;
}



/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_markup                                               */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_markup( const char *f, char d )

{
  char buffer[1000];
  char *l=label, *r=buffer;
  for(;;) {
    if (*f && (*l == 0 || slash_feature_boundary(l)))
      {
	*(r++) = d;
	while (*f)
	  *(r++) = *(f++) ;
	while ((*(r++) = *(l++))) ;
	break;
      }
    if ((*(r++) = *(l++)) == 0)
      break;
  }
  free(label);
  label = strdup(buffer);
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::insert_feature                                           */
/*                                                                 */
/*FE****************************************************************/

void Tree::insert_feature( const char *f )

{
  add_markup(f, '/');
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::insert_uniq_feature                                      */
/*                                                                 */
/*FE****************************************************************/

void Tree::insert_uniq_feature( const char *f )

{
  for( char *p=label; *p; p++ )
    if (*p == '/')
      for( int i=0; ; i++ )
	if (f[i] == 0 && (p[i+1] == 0 || feature_boundary(p+1)))
	  return; // feature was previously added
	else if (f[i] != p[i+1])
	  break;
  
  add_markup(f, '/');
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::cut_off_indices                                          */
/*                                                                 */
/*FE****************************************************************/

const char *Tree::cut_off_indices()

{
  for( char *p=label; *p; p++ )
    if ((*p == '-'  || *p == '=') && is_digit(*(p+1))) {
      char *e;
      for( e=p+2; *e; e++ )
	if (is_digit(*e))
	  ;
	else if ((*e == '-' || *e == '=') && is_digit(*(e+1)))
	  e++;
	else
	  break;
      if (*e == 0) {
	static char buffer[1000];
	for( int i=0; (buffer[i] = p[i]); i++ );
	*p = 0;
	return buffer;
      }
    }
  return "";
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::append_slash                                             */
/*                                                                 */
/*FE****************************************************************/

void Tree::append_slash( char *prefix, char c, bool terminal )

{
  char buffer[1000];
  const char *indices=cut_off_indices();
  if (terminal)
    sprintf(buffer,"%s%c#%s%s",label,c,prefix,indices);
  else
    sprintf(buffer,"%s%c%s%s",label,c,prefix,indices);
  change_label(buffer);
}



/*FA****************************************************************/
/*                                                                 */
/*  quote                                                          */
/*                                                                 */
/*FE****************************************************************/

static char *quote( const char *string )

{
  int i,k;
  static char buffer[10000];
  bool is_trace=false;

  if (string[0] == '(' && *last_char(string) == ')')
    is_trace = true;
  
  for( i=k=0; k<10000; i++, k++) {
    if (!string[i])
      break;
    if (string[i] == '"' || string[i] == '\\'|| string[i] == ' ' || 
	string[i] == '$' || string[i] == '^' || string[i] == '\''||
	string[i] == '[' || string[i] == ']' || string[i] == '{' ||
	string[i] == '}' || string[i] == '=' || string[i] == '<' || 
	string[i] == '>' || string[i] == '#' || string[i] == '/' || 
	(!is_trace && (string[i] == '(' || string[i] == ')')))
      buffer[k++] = '\\';
    buffer[k] = string[i];
  }
  buffer[k] = '\0';

  return &buffer[0];
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::print_subtree                                            */
/*                                                                 */
/*FE****************************************************************/

void Tree::print_subtree( int depth )

{
  if (daughters.size() == 0)
    fprintf(outfile," %s", quote(label));
  else {
    if (PrettyPrint) {
      fputc('\n', outfile);
      for( int i=0; i<depth; i++ )
	fputs("  ", outfile);
    }
    if (WithHeads && mother && this == mother->head_daughter)
      fprintf(outfile, "(%s", quote(label));
    else
      fprintf(outfile, "(%s", quote(label));
    for( size_t i=0; i<daughters.size(); i++ )
      daughters[i]->print_subtree(depth+1);
    fputc(')', outfile);
  }
}

/*FA****************************************************************/
/*                                                                 */
/*  Tree::print_unary_branches                                     */
/*                                                                 */
/*FE****************************************************************/

bool Tree::print_unary_branches()

{
  bool result = false;

  if (mother && daughters.size() == 1 && daughters[0]->daughters.size() > 0) {
    printf("(%s %s)\n", label, daughters[0]->label);
    result = true;
  }

  for( size_t i=0; i<daughters.size(); i++ )
    if (daughters[i]->print_unary_branches())
      result = true;

  return result;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::print                                                    */
/*                                                                 */
/*FE****************************************************************/

void Tree::print()

{
  print_subtree(0);
  fputc('\n', outfile);
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::matching_trace                                           */
/*                                                                 */
/*FE****************************************************************/

char Tree::matching_trace( const char *suffix )

{
  // find trace index
  for( char *p=label+1; *p; p++ )
    // start of a trace index?
    if (*p == '-' || *p == '=')
      if (daughters.size() > 0 || *(p-1) == '*') {
	
	// match the following characters with the suffix
	const char *s=suffix;
	for( char *pp=p+1; ; pp++, s++ ) {
	  if (*s == 0) {
	    // matching complete
	    if (*pp == 0 || *pp == '-' || *pp == '=') {
	      char c;
	      if (*p == '-')
		c = '_';
	      else c = *p;
	      // delete the trace index
	      while ((*(p++) = *(pp++))) ;
	      return c;
	    }
	    break;
	  }
	  else if (*pp != *s)
	    break; // mismatch
	}
      }
  return 0;
}


/*FA****************************************************************/
/*                                                                 */
/*  split_name                                                     */
/*                                                                 */
/*FE****************************************************************/

vector<char*> &split_name( char *s, size_t nmax )

{
  static vector<char*> result;
  for( size_t i=0; i<result.size(); i++ )
    free(result[i]);
  result.clear();

  for( size_t i=0; i<nmax; i++ ) {
    char buffer[1000];
    char *s1=buffer;
    *(s1++) = *(s++);
    while (*s && ((*s != '-' && *s != '/') || i == nmax-1))
      *(s1++) = *(s++);
    *s1 = 0;
    result.push_back(strdup(buffer));
    if (*s == 0)
      break;
    if (*s != '-')
      i = nmax-2;
  }
  while (result.size() < nmax)
    result.push_back(strdup(""));

  return result;
}


/*FA****************************************************************/
/*                                                                 */
/*  skip_over_features                                             */
/*                                                                 */
/*FE****************************************************************/

void skip_over_features( char* &s )

{
  while (s[0] == '.' || s[0] == '-') {
    int l=0;
    if (strpref(s+1,"Subj") || strpref(s+1,"Pres") || 
	strpref(s+1,"Past") || strpref(s+1,"Masc") || strpref(s+1,"Neut"))
      l = 5;
    else if (strpref(s+1,"Psp") || strpref(s+1,"Ind") || strpref(s+1,"Fem"))
      l = 4;
    else if (strpref(s+1,"Sg") || strpref(s+1,"Pl"))
      l = 3;
    else if (s[1] >= '1' && s[1] <= '3' && s[2] == '.')
      l = 2;
    else if (s[1] == '*')
      l = 2;
    if (l > 0 && (s[l] == '.' || s[l] == '-' || s[l] == 0))
      s += l;
    else
      return;
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  only_number_follows                                            */
/*                                                                 */
/*FE****************************************************************/

static bool only_number_follows( char *s )

{
  while (*s >= '0' && *s <= '9')
    s++;
  return *s == 0;
}



/*FA****************************************************************/
/*                                                                 */
/*  eliminate_agr_features                                         */
/*                                                                 */
/*FE****************************************************************/

static void eliminate_agr_features( char *label )

{
  int level=0;
  if (strncmp(label,"VVIMP",5) == 0)
    level = 1;

  // scan the POS tag and the semantic tag
  char *p;
  if ((p = index( label, '-')) == NULL)
    return; // no agreement features
  if ((p = index( p+1, '-')) == NULL)
    return; // no agreement features
  if (only_number_follows(p+1))
    return; // no agreement features

  char *id = index( p+1, '-');
  if (id)
    *(id++) = 0;

  char *r=p+1;
  do {
    char *f = p+1;
    p = index(f, '.');
    if (p)
      *p = 0;

    if (FeatureClass[f] > level) {
      // retain this feature
      while (*f)
	*(r++) = *(f++);
      *(r++) = '.';
    }
  }
  while (p);

  if (id) {
    *(r-1) = '-';
    while ((*(r++) = *(id++)));
  }
  else
    *(r-1) = 0;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::delete_features                                          */
/*                                                                 */
/*FE****************************************************************/

void Tree::delete_features()

{
  if (daughters.size() == 0)
    return;

  if (daughters[0]->daughters.size() == 0)
    eliminate_agr_features( label );
  
  if (matching_base_cat("TRUNC") && daughters[0]->label[0] == '-') {
    char buffer[100];
    sprintf(buffer,"NN%s", label+5);
    free(label);
    label = strdup(buffer);
  }

  if (daughters.size() == 0 && is_ordinal(label)) {
    vector<char*> &s=split_name(mother->label, 3);
    free(mother->label);
    char buffer[100];
    sprintf(buffer,"ORD%s", s[1]);
    mother->label = strdup(buffer);
  }

  if (daughters.size() == 1 && matching_base_cat("TRUNC")) {
    char c=daughters[0]->label[0];
    if (c != '-') {
      vector<char*> &s=split_name(label, 3);
      free(label);
      char buffer[100];
      if ((c >= 'A' && c <= 'Z') || (c >= '' && c<= ''))
	sprintf(buffer,"TRUNCCap%s", s[1]);
      else if ((c >= 'a' && c <= 'z') || (c >= '' && c<= ''))
	sprintf(buffer,"TRUNCLC%s", s[1]);
      else
	sprintf(buffer,"TRUNC%s", s[1]);
      label = strdup(buffer);
    }
  }

  // recursion
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->delete_features();
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::map_categories                                           */
/*                                                                 */
/*FE****************************************************************/

void Tree::map_categories()

{
  if (daughters.size() == 0)
    return;

  bool modify = false;
  vector<char*> &s=split_name(label, 3);

  // Insert NONE relation
  if (mother && strcmp(s[1], "") == 0 && label[0] != '$') { // node without GF
    free(s[1]);
    s[1] = strdup("-NONE");
    modify = true;
  }

  // simplify the categories
  char buffer[100];
  sprintf(buffer,"%s%s",s[0], s[1]);
  const char *l = CatMap[buffer];

  if (l) {
    free(s[0]);
    free(s[1]);
    s[0] = strdup(l);
    s[1] = strdup("");
    modify = true;
  }

  if (modify) {
    char buffer[1000];
    sprintf(buffer,"%s%s%s", s[0], s[1], s[2]);
    change_label(buffer);
  }

  // recursion
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->map_categories();
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::convert_conjunctions                                     */
/*                                                                 */
/*FE****************************************************************/

void Tree::convert_conjunctions()

{
  if (daughters.size() == 0)
    return;

  if (label[0] == 'C') {
    char buffer[1000], rel[100];
    vector<char*> &s = split_name(label+1, 3);
    sprintf(rel,"%s",s[1]);
    for( size_t i=0; i<daughters.size(); i++ ) {
      s = split_name(daughters[i]->label, 3);
      if (strcmp(s[1],"-CJ") == 0) {
	sprintf(buffer,"%s%s%s", s[0], rel, s[2]);
	daughters[i]->change_label(buffer);
      }
    }
  }

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->convert_conjunctions();
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::insert_chain_node                                        */
/*                                                                 */
/*FE****************************************************************/

void Tree::insert_chain_node( const char *cat, size_t pos )

{
  Tree *d=daughters[pos];

  // create the new node
  char buffer[1000];
  vector<char*> &s=split_name( d->label, 4 );
  if (s[3][0] == 0 && s[2][0] == '-' && is_digit(s[2][1]) && s[2][2] == 0) {
    char *tmp = s[2];
    s[2] = s[3];
    s[3] = tmp;
  }
  sprintf(buffer,"%s-HD%s", s[0], s[2]);
  free( d->label );
  d->label = strdup(buffer);

  if (label[0] == 'C')
    sprintf(buffer,"%s%s%s/chain", cat, s[1], s[3]);
  else
    sprintf(buffer,"%s%s%s", cat, s[1], s[3]);
  Tree *node=new Tree(strdup(buffer));
  daughters[pos] = node;
  node->mother = this;
  node->daughters.push_back(d);
  d->mother = node;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::insert_chain_rules                                       */
/*                                                                 */
/*FE****************************************************************/

void Tree::insert_chain_rules()

{
  // recursion
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->insert_chain_rules();

  if (matching_base_cat("S-RC"))
    for( size_t i=0; i<daughters.size(); i++ )
      if (daughters[i]->matching_base_cat("PRELS"))
	insert_chain_node( "NP", i );

  if (matching_base_cat("S") || matching_base_cat("VP") || 
      matching_base_cat("TOP") || matching_base_cat("DL"))
    for( size_t i=0; i<daughters.size(); i++ ) {
      Tree *d=daughters[i];
      if (d->matching_base_cat("NE"))
	insert_chain_node( "PN", i );

      else if (d->matching_base_cat("NN") || d->matching_base_cat("PDS") || 
	       d->matching_base_cat("PPER") || d->matching_base_cat("PIS") ||
	       d->matching_base_cat("PN") || d->matching_base_cat("CARD"))
	insert_chain_node( "NP", i );
      
    }
  
  if (matching_base_cat("NP") || matching_base_cat("CNP") || 
      matching_base_cat("CO") || matching_base_cat("PP") ||
      matching_base_cat("AP"))
    for( size_t i=0; i<daughters.size(); i++ ) {
      Tree *d=daughters[i];
      if (d->matching_base_cat("NN") &&
	  (d->matching_base_cat("NN-SB") ||
	   d->matching_base_cat("NN-OA") ||
	   d->matching_base_cat("NN-DA") ||
	   d->matching_base_cat("NN-OG") ||
	   d->matching_base_cat("NN-APP") ||
	   d->matching_base_cat("NN-AG") ||
	   d->matching_base_cat("NN-PAR") ||
	   d->matching_base_cat("NN-MNR") ||
	   d->matching_base_cat("NN-OP") ||
	   d->matching_base_cat("NN-RE") ||
	   d->matching_base_cat("NN-MO") ||
	   d->matching_base_cat("NN-TOP") ||
	   d->matching_base_cat("NN-PH") ||
	   d->matching_base_cat("NN-PD") ||
	   d->matching_base_cat("NN-VO")))
	insert_chain_node( "NP", i );
      else if (d->matching_base_cat("NE") &&
	       (d->matching_base_cat("NE-SB") ||
		d->matching_base_cat("NE-OA") ||
		d->matching_base_cat("NE-DA") ||
		d->matching_base_cat("NE-OG") ||
		d->matching_base_cat("NE-AG") ||
		d->matching_base_cat("NE-APP") ||
		d->matching_base_cat("NE-PAR") ||
		d->matching_base_cat("NE-OP") ||
		d->matching_base_cat("NE-MNR") ||
		d->matching_base_cat("NE-MO") ||
		d->matching_base_cat("NE-RE") ||
		d->matching_base_cat("NE-TOP") ||
		d->matching_base_cat("NE-VO")))
	insert_chain_node( "PN", i );
    }
}


/*FA****************************************************************/
/*                                                                 */
/*  trace_category                                                 */
/*                                                                 */
/*FE****************************************************************/

static char *trace_category( char *label )

{
  static char buffer[1000];
  char *p=buffer;
  bool copy=true;
  for( ; *label; label++ ) {
    if (*label == '/')
      copy = false;
    else if (*label == '-') {
      if (strcmp(buffer,"NP") == 0 || strcmp(buffer,"S") == 0)
	label++;
      else
	copy = false;
    }
    else if (*label == '=' || *label == '%' || *label == '+' || 
	     index_start(label))
      break;
    else if (*label == '_')
      copy = true;
    if (copy) {
      *(p++) = *label;
      *p = 0;
    }
  }
  *p = 0;
  return buffer;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::append_slashes                                           */
/*                                                                 */
/*FE****************************************************************/

char Tree::append_slashes( char *prefix, const char *suffix,
			   bool &terminal )
{
  // current node carries the index?
  char c;
  if ((c=matching_trace(suffix)))
    terminal = true;
  else {
    terminal = false;
    // look for the index in the subtree
    for( int i=(int)daughters.size()-1; i>=0; i-- ) {
      bool term;
      if ((c = daughters[i]->append_slashes(prefix, suffix, term))) {
	// match in subtree
	daughters[i]->append_slash( prefix, c, term );
	break;
      }
    }
  }

  return c;
}



/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_slashes                                              */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_slashes()

{
  if (daughters.size() == 0)
    return;

  // process daughter nodes
  for( int i=(int)daughters.size()-1; i>=0; i-- )
    daughters[i]->add_slashes();

  // Is current node a filler?
  char *s;
  while ((s = find_index(label)) != NULL) {
    *(s++) = 0; // delete the index
    char *p=trace_category(label);
    bool match_found = false;
    s = strdup(s);
    for(;;){
      Tree *t;
      char c=0;
      bool term; // not needed actually
      for( t=this; t; t=t->mother )
	if ((c = t->append_slashes( p, s, term )) != 0)
	  break;
      if (t == NULL)
	break; // exit loop

      match_found = true;
      if (t == this)
	append_slash( p, '+', true );
      else {
	if (c == '_')
	  c = '+';
	else
	  c = '%';
	append_slash( p, c, true );
	for( Tree *t2=mother; t2 != t; t2=t2->mother )
	  t2->append_slash( p, c, false );
      }
    }
    if (!match_found)
      fprintf(stderr,"\tWarning: no matching trace for \"%s-%s\"!\n",label,s);
    free(s);
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::matching_prefix                                          */
/*                                                                 */
/*FE****************************************************************/

bool Tree::matching_prefix( const char *prefix )

{
  for( char *p=label; *prefix; p++, prefix++ )
    if (*p != *prefix)
      return false;
  return true;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::has_trace_feature                                        */
/*                                                                 */
/*FE****************************************************************/

bool Tree::has_trace_feature( char d, const char *s )

{
  for( const char *p=label; *p; p++ )
    if (*p == d) {
      const char *p1=p+1;
      if (*p1 == '#')
	p1++;
      if (*p1 == '') {
	const char *p2=s;
	p1++;
	// compare the categories
	while (*p2 && *(p1++) == *(p2++)) ;
	if (*p1 == '')
	  return true;
      }
  }
  return false;
}



/*FA****************************************************************/
/*                                                                 */
/*  Tree::mark_as_trace                                            */
/*                                                                 */
/*FE****************************************************************/

void Tree::mark_as_trace()

{
  char *buffer = (char*)malloc(strlen(label) + 3);
  sprintf(buffer, "(%s)", label);
  free(label);
  label = buffer;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::collapse_traces                                          */
/*                                                                 */
/*FE****************************************************************/

bool Tree::collapse_traces()

{
  char buffer[10000];
  char *p=buffer;

  if (daughters.size() == 0)
    return matching_category("*T*");

  p += sprintf(p,"%s",label);
  bool is_trace=true;
  for( size_t i=0; i<daughters.size(); i++ )
    if (daughters[i]->collapse_traces()) {
      p += sprintf(p,"(%s)",daughters[i]->label);
      daughters[i]->mark_as_trace();
    }
    else
      is_trace = false;
  if (!is_trace)
    return false;

  change_label(buffer);
  for( size_t i=0; i<daughters.size(); i++ )
    delete daughters[i];
  daughters.clear();

  return true;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::delete_indices                                           */
/*                                                                 */
/*FE****************************************************************/

void Tree::delete_indices()

{
  if (daughters.size() > 0 || matching_category("*T*")) {
    cut_off_indices();
    for( size_t i=0; i<daughters.size(); i++ )
      daughters[i]->delete_indices();
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::delete_traces                                            */
/*                                                                 */
/*FE****************************************************************/

bool Tree::delete_traces()

{
  if (daughters.size() == 0)
    return is_trace();

  bool is_trace = true;
  size_t k=0;
  for( size_t i=0; i<daughters.size(); i++ )
    if (daughters[i]->delete_traces())
      delete daughters[i];
    else {
      daughters[k++] = daughters[i];
      is_trace = false;
    }

  daughters.resize(k);
  return is_trace;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::delete_chains                                            */
/*                                                                 */
/*FE****************************************************************/

void Tree::delete_chains()

{
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->delete_chains();

  if (daughters.size() > 0 && matching_base_cat("ORD")) {
    char buffer[100];
    vector<char*> &s=split_name( label, 2 );
    sprintf(buffer,"ADJA%s", s[1]);
    free(label);
    label = strdup(buffer);
  }

  if (mother && daughters.size() == 1 && daughters[0]->daughters.size() > 0)
    if (matching_base_cat("NP") || 
	matching_base_cat("CNP") || 
	matching_base_cat("AP") || 
	matching_base_cat("AVP") || 
	(matching_base_cat("PN") && !daughters[0]->matching_base_cat("NP")) ||
	(matching_base_cat("PP") && daughters[0]->matching_base_cat("PROAV-PH")) ||
	(matching_base_cat("") && daughters[0]->matching_base_cat("")))
      {
	char rel[100];
	vector<char*> &s=split_name( label, 3 );
	sprintf(rel,"%s", s[1]);
	vector<char*> &c=split_name( daughters[0]->label, 3 );
	
	// create a new label
	char buffer[100];
	sprintf(buffer,"%s%s%s", c[0], rel, c[2]);
	free(label);
	label = strdup(buffer);
	
	// remove the node
	Tree *tmp = daughters[0];
	daughters = tmp->daughters;
	tmp->daughters.clear();
	delete tmp;
      }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::print_govs                                               */
/*                                                                 */
/*FE****************************************************************/

void Tree::print_govs()

{
  if (daughters.size() == 0)
    return;

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->print_govs();

  if (mother && head != mother->head) {
    if (head)
      printf("%s\t%s\t", head->daughters[0]->label, head->label);
    else
      printf("<NONE>\t-NONE-\t");
    
    if (mother->head2 && mother->head2 != head)
      printf( "%s\t%s\t", mother->head2->daughters[0]->label, 
	      mother->head->label);
    else if (mother->head)
      printf( "%s\t%s\t", mother->head->daughters[0]->label, 
	      mother->head->label);
    else
      printf( "<NONE>\t-NONE-\t");
    printf( "%s\n", mother->label);
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::aux_label                                                */
/*                                                                 */
/*FE****************************************************************/

static bool aux_label( char *s )

{
  if (*(s++) != '<')
    return false;

  while (*s != '[') {
    if (!*s)
      return false;
    s++;
  }

  while (*s != ']') {
    if (!*s)
      return false;
    s++;
  }

  while (*s != '>') {
    if (!*s)
      return false;
    s++;
  }

  return *(s+1) == 0;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::undo_markov                                              */
/*                                                                 */
/*FE****************************************************************/

void Tree::undo_markov()

{
  if (daughters.size() > 0) {
    Trees d;
    
    for( size_t i=0; i<daughters.size(); i++ ) {
      Tree *daughter=daughters[i];
      daughter->undo_markov();
      if (aux_label(daughter->label)) {
	for( size_t k=0; k<daughter->daughters.size(); k++ )
	  d.push_back(daughter->daughters[k]);
	daughter->daughters.clear();
	delete daughter;
      }
      else
	d.push_back(daughter);
    }
    daughters = d;
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::restore_gf                                               */
/*                                                                 */
/*FE****************************************************************/

void Tree::restore_gf()

{
  if (daughters.size() == 0)
    return;

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->restore_gf();

  if (matching_base_cat("NP") || matching_base_cat("PP")) {
    for( size_t i=0; i<daughters.size(); i++ ) {
      vector<char*> &s=split_name(daughters[i]->label, 3);
      if (strcmp(s[1],"-HD") == 0)
	daughters[i]->change_gf_label("-NK");
    }
  }
  else {
    vector<char*> &s = split_name(label, 3);
    if (label[0] == 'C') {
      char *gf = strdup(s[1]);
      for( size_t i=0; i<daughters.size(); i++ ) {
	s = split_name(daughters[i]->label, 3);
	if (strcmp(s[1], gf) == 0) {
	  daughters[i]->change_gf_label("-CJ");
	}
      }
      free(gf);
    }
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_domV_feature                                         */
/*                                                                 */
/*FE****************************************************************/

bool Tree::add_domV_feature()

{
  bool result=false;
  for( size_t i=0; i<daughters.size(); i++ )
    if (daughters[i]->add_domV_feature())
      result = true;

  if (matching_base_cat("VP") || matching_base_cat("S"))
    return true;

  if (mother && result)
    insert_feature("domV");

  return result;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_adverb_features                                      */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_adverb_features()

{
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_adverb_features();

  if (matching_base_cat("AVP"))
    insert_feature(mother->base_cat());
  else if (matching_base_cat("ADV") || matching_base_cat("ADJD")) {
    if (mother->matching_base_cat("AVP"))
      insert_feature(mother->mother->base_cat());
    else if (mother->matching_base_cat("AP")) {
      const char *f = mother->base_cat();
      for( size_t i=0; i<mother->daughters.size(); i++ )
	if (mother->daughters[i]->matching_base_cat("NM-HD") ||
	    mother->daughters[i]->matching_base_cat("CARD-HD"))
	  {
	    f = "card";
	    break;
	  }
      insert_feature(f);
    }
    else
      insert_feature(mother->base_cat());
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_punctuation_features                                 */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_punctuation_features()

{
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_punctuation_features();

  if (matching_base_cat("S-TOP")) {
    Tree *punc=NULL;
    for( size_t i=0; i<mother->daughters.size(); i++ )
      if (strcmp(mother->daughters[i]->label, "$.") == 0)
	punc = mother->daughters[i];
    if (punc) {
      if (strcmp(punc->daughters[0]->label, "!") == 0) {
	insert_feature("imp");
	punc->insert_feature("imp");
      }
      else if (strcmp(punc->daughters[0]->label, "?") == 0) {
	insert_feature("quest");
	punc->insert_feature("quest");
      }
    }
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::correct_labels                                           */
/*                                                                 */
/*FE****************************************************************/

void Tree::correct_labels()

{
  if (daughters.size() == 0) {
    if (strcmp(label,"-") == 0 && mother->matching_base_cat("APPR"))
      mother->change_label("$Par");
    if (strcmp(label,"von") == 0 && mother->matching_base_cat("NE-PNC"))
      mother->change_label("APPR-PNC");

    return;
  }
  
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->correct_labels();

  if (matching_category("NN-AG-Nom"))
    memcpy(label+6, "Gen", 3);
  else if (matching_base_cat("NN-PNC"))
    label[1] = 'E';

  if (matching_base_cat("S-RC") && daughters[0]->matching_base_cat("ART"))
    daughters[0]->change_base_label("PRELS");

  char *p = index(label, '-');
  if (p && strpref(p,"-NK")) {
    *(p+1) = 'H';
    *(p+2) = 'D';
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_rel_features                                         */
/*                                                                 */
/*FE****************************************************************/

bool Tree::add_rel_features()

{
  size_t i;
  for( i=0; i<daughters.size(); i++ )
    if (daughters[i]->label[0] != '$' &&
	!daughters[i]->matching_base_cat("APPR"))
      break;

  if (i == daughters.size())
    return false;

  if (daughters[i]->matching_base_cat("NP") ||
      daughters[i]->matching_base_cat("PP") ||
      daughters[i]->matching_base_cat("VP") ||
      daughters[i]->matching_base_cat("CVP"))
    {
      if (daughters[i]->add_rel_features()) {
	daughters[i]->insert_feature("rel");
	return true;
      }
      return false;
    }

  if (daughters[i]->matching_base_cat("PRELS") ||
      daughters[i]->matching_base_cat("PRELAT"))
    return true;


  if (daughters[i]->matching_base_cat("PROAV") ||
      daughters[i]->matching_base_cat("PWAV") ||
      daughters[i]->matching_base_cat("PWS") ||
      daughters[i]->matching_base_cat("KOUS") ||
      daughters[i]->matching_base_cat("PDS-AG"))
    {
      daughters[i]->insert_feature("rel");
      return true;
    }

  return false;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_NP_features                                          */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_NP_features()

{
  if (daughters.size() == 0)
    return;

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_NP_features();

  if (matching_base_cat("S-RC"))
    if (!add_rel_features())
      insert_feature("norel");

  if ((matching_base_cat("PWAT") || matching_base_cat("PWAV") ||
       matching_base_cat("PWS")) &&
      (mother->matching_base_cat("NP") || mother->matching_base_cat("PP")))
    mother->insert_feature("wh");

  if (matching_base_cat("NP-SB")) {
    for( size_t i=0; i<daughters.size(); i++ ) {
      vector<char*> &s=split_name(daughters[i]->label, 3);
      if (strstr(s[2],"Sg")) {
	insert_feature("Sg");
	break;
      }
      else if (strstr(s[2],"Pl")) {
	insert_feature("Pl");
	break;
      }
      else if (strcmp(s[0],"CARD") == 0) {
	insert_feature("Pl");
	break;
      }
    }
  }

  if (matching_base_cat("PPER") && mother->matching_base_cat("NP-EP"))
    insert_feature("ep");

  if (matching_base_cat("NN") && 
      (mother->matching_base_cat("NP") || mother->matching_base_cat("PP")))
    {
      Tree *t = successor();
      if (t) {
	if (t->matching_base_cat("NN-HD"))
	  insert_feature("seq");
	else if (t->matching_base_cat("$Par") &&
		 strcmp(t->daughters[0]->label, "*LRB*") != 0) 
	  {
	    Tree *t2 = t->successor();
	    if (t2 && t2->matching_base_cat("NN-HD"))
	      insert_feature("seq");
	  }
      }
    }

  if (matching_base_cat("PN")) {
    char buffer[100];
    sprintf(buffer,"%s", label);
    for( size_t i=0; i<daughters.size(); i++ )
      if (daughters[i]->matching_base_cat("NE")) {
	vector<char*> &s=split_name(daughters[i]->label, 3);
	sprintf(buffer,"%s%s", label, s[2]);
      }
    free(label);
    label = strdup( buffer );
  }

  else if (matching_base_cat("NP-PNC"))
    for( size_t i=0; i<daughters.size(); i++ ) {
      if (daughters[i]->matching_base_cat("NN") ||
	  daughters[i]->matching_base_cat("ADJA"))
	daughters[i]->insert_feature("name");
    }
}



/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_PP_features                                          */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_PP_features()

{
  if (daughters.size() == 0)
    return;

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_PP_features();

  if (matching_base_cat("PP")) {
    if (mother->matching_base_cat("CPP")) {
      mother->insert_uniq_feature(mother->mother->base_cat());
      insert_feature(mother->mother->base_cat());
    }
    else
      insert_feature(mother->base_cat());
  }

  if (matching_base_cat("APPR") && !matching_category("APPR-PNC")) {
    char c[4];
    c[0] = c[3] = 0;
    for( size_t i=0; i<mother->daughters.size(); i++) {
      Tree *node = mother->daughters[i];
      vector<char*> &s=split_name( node->label, 3 );
      if (strpref(s[2],"-Nom") || strpref(s[2],"-Gen") || 
	  strpref(s[2],"-Dat") || strpref(s[2],"-Acc"))
	strncpy(c, s[2]+1, 3);
      else if (node->matching_base_cat("CNP")) {
	for( size_t i=0; i<node->daughters.size(); i++) {
	  vector<char*> &s=split_name( node->daughters[i]->label, 3 );
	  if (strpref(s[2],"-Nom") || strpref(s[2],"-Gen") || 
	      strpref(s[2],"-Dat") || strpref(s[2],"-Acc"))
	    {
	      strncpy(c, s[2]+1, 3);
	      node->insert_feature( c );
	      break;
	    }
	}
      }
    }
    if (c[0] != 0)
      insert_feature(c);
  }

  if (strpref(label,"APPR") && !matching_category("APPR-PNC")) {
    char *l=lowercase(daughters[0]->label);
    if (PrepMap.find(l) != PrepMap.end())
      insert_feature(PrepMap[l]);
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_AP_features                                          */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_AP_features()

{
  if (daughters.size() == 0)
    return;

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_AP_features();

  if (matching_base_cat("AP") || matching_base_cat("CAP"))
    for( size_t i=0; i<daughters.size(); i++ )
      if (daughters[i]->matching_base_cat("ADJA-HD") ||
	  daughters[i]->matching_base_cat("AP-HD") ||
	  daughters[i]->matching_base_cat("CAP-HD") ||
	  daughters[i]->matching_base_cat("PIAT-HD") ||
	  daughters[i]->matching_base_cat("PDAT-HD") ||
	  daughters[i]->matching_base_cat("ART-HD"))
	{
	  vector<char*> &s=split_name( daughters[i]->label, 3 );
	  if (s[2][0] != 0) {
	    add_markup(s[2]+1, '-');
	    break;
	  }
	}
      else if (daughters[i]->matching_base_cat("ADJD-HD") ||
	       daughters[i]->matching_category("AP-PD/pred") ||
	       daughters[i]->matching_category("AP-HD/pred") ||
	       daughters[i]->matching_category("AP-TOP/pred"))
	{
	  insert_feature("pred");
	  break;
	}
      else if (daughters[i]->matching_base_cat("NN-HD")) {
	vector<char*> &s=split_name( daughters[i]->label, 3 );
	if (s[2][0] != 0) {
	  add_markup(s[2]+1, '-');
	  insert_feature("nom");
	  daughters[i]->insert_feature("adj");
	  break;
	}
      }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_CARD_features                                        */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_CARD_features()

{
  if (daughters.size() == 0)
    return;

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_CARD_features();

  if (matching_base_cat("CARD") && is_year_number(daughters[0]->label))
    insert_feature("year");
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_KOUS_features                                        */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_KOUS_features()

{
  if (daughters.size() == 0)
    return;

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_KOUS_features();

  if (matching_category("KOUS-CP")) {
    if (mother->matching_category("S-MO"))
      insert_feature("mo");
    else if (mother->matching_category("S-OC"))
      insert_feature("oc");
    else if (mother->matching_category("S-SB"))
      insert_feature("sb");
    else if (mother->matching_category("S-RE"))
      insert_feature("re");
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_KON_features                                         */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_KON_features()

{
  if (daughters.size() == 0)
    return;

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_KON_features();

  // error corrections
  if (matching_category("KON-CD-Dat") || matching_category("KON-CD-Acc"))
    change_label("KON-CD");
  if (matching_category("ADV-MO") && 
      strcmp(daughters[0]->label,"entweder") == 0)
    {
      change_label("KON-CD");
    }

  if (matching_category("KON-CD")) {
    char *word = lowercase(daughters[0]->label);
    if (strcmp(word,"sowohl") == 0)
      insert_feature("sowohl");
    else if (strcmp(word,"als") == 0)
      insert_feature("als");
    else if (strcmp(word,"weder") == 0)
      insert_feature("weder");
    else if (strcmp(word,"noch") == 0)
      insert_feature("noch");
    else if (strcmp(word,"entweder") == 0)
      insert_feature("entweder");
    else if (strcmp(word,"oder") == 0) {
      size_t i;
      for( i=0; i<mother->daughters.size(); i++ )
	if (mother->daughters[i]->matching_category("KON-CD") &&
	    strcmp(lowercase(mother->daughters[i]->daughters[0]->label), 
		   "entweder") == 0)
	  {
	    insert_feature("oder");
	    break;
	  }
    }
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_VP_features                                          */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_VP_features()

{
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_VP_features();

  if (matching_base_cat("VP-OC")) {
    int i;
    for( i=(int)daughters.size()-1; i>=0; i-- ) {
      if (daughters[i]->matching_base_cat("V?PP-HD"))
	insert_feature("pp");
      else if (daughters[i]->matching_base_cat("V?INF-HD"))
	insert_feature("inf");
      else if (daughters[i]->matching_base_cat("V?IMP-HD"))
	insert_feature("imp");
      else if (daughters[i]->matching_base_cat("V?FIN-HD")) {
	insert_feature("fin");
	vector<char*> &s=split_name(daughters[i]->label, 3);
	if (strstr(s[2],"Sg"))
	  insert_feature("Sg");
	else if (strstr(s[2],"Pl"))
	  insert_feature("Pl");
      }
      else if (daughters[i]->matching_base_cat("V?IZU-HD"))
	insert_feature("zu");
      else if (daughters[i]->matching_base_cat("VZ-HD"))
	insert_feature("zu");
      else
	continue;
      break;
    }
  }

  else if (matching_base_cat("CVP-OC"))
    for( size_t i=0; i<daughters.size(); i++ )
      if (daughters[i]->matching_base_cat("VP-OC")) {
	insert_feature(daughters[i]->label+6);
	break;
      }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_nohead_feature                                       */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_nohead_feature()

{
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_nohead_feature();

  if (mother == NULL || daughters.size() == 0 || 
      matching_base_cat("ADV-MO") || matching_base_cat("ADJD-MO") || 
      (daughters.size() == 1 && daughters[0]->daughters.size() == 0))
    return;

  size_t  i;
  for( i=0; i<daughters.size(); i++ ) {
    vector<char*> &s=split_name( daughters[i]->label, 3 );
    if (strcmp(s[1],"-HD") == 0  || strcmp(s[1],"-NK") == 0 ||
	strcmp(s[1],"-PNC") == 0 || strcmp(s[1],"-CJ") == 0 ||
	strcmp(s[1],"-AC") == 0  || strcmp(s[1],"-AVC") == 0 ||
	strcmp(s[1],"-NMC") == 0 || strcmp(s[1],"-PH") == 0 ||
	strcmp(s[1],"-PD") == 0  || strcmp(s[1],"-ADC") == 0 ||
	strcmp(s[1],"-UC") == 0  || strcmp(s[1],"-DH") == 0)
      break;
  }
  if (i == daughters.size() && label[0] != 'C')
    // no head and not a conjunction
    insert_feature("nohead");
}



/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_nosubj_feature                                       */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_nosubj_feature()

{
  if (daughters.size() == 0)
    return;

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_nosubj_feature();

  if (matching_base_cat("S") && mother->matching_base_cat("CS")) {
    for( size_t i=0; i<daughters.size(); i++ ) {
      if (daughters[i]->matching_base_cat("VVIMP"))
	return;
      vector<char*> &s=split_name(daughters[i]->label, 3);
      if (strcmp(s[1], "-SB") == 0 || strcmp(s[1], "-EP") == 0)
	return;
    }
    insert_feature("nosubj");
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_subcat_feature                                       */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_subcat_feature()

{
  if (daughters.size() == 0)
    return;

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_subcat_feature();

  if (matching_base_cat("VP") || matching_base_cat("S")) {
    for( size_t i=0; i<daughters.size(); i++ )
      if (daughters[i]->matching_base_cat("VP"))
	return;
      else if (strpref(daughters[i]->label,"VVFIN") ||
	       strpref(daughters[i]->label,"VVINF")) {
	char features[1000], *p=features;
	*p = '0';
	*(p+1) = 0;
	for( size_t k=0; k<daughters.size(); k++ ) {
	  if (daughters[k]->matching_base_cat("NP-OA") ||
	      daughters[k]->matching_base_cat("CNP-OA"))
	    append_string(p, "a");
	  else if (daughters[k]->matching_base_cat("NP-DA") ||
		   daughters[k]->matching_base_cat("CNP-DA"))
	    append_string(p, "d");
	  else if (daughters[k]->matching_base_cat("PTKVZ-SVP"))
	    append_string(p, "t");
	  else if (daughters[k]->matching_base_cat("PRF-OA"))
	    append_string(p, "r");
	  else if (daughters[k]->matching_base_cat("PP-OP") ||
		   daughters[k]->matching_base_cat("CPP-OP"))
	    append_string(p, "p");
	  else if (daughters[k]->matching_base_cat("NP-EP"))
	    append_string(p, "e");
	  else if (daughters[k]->matching_base_cat("PP-CVC") ||
		   daughters[k]->matching_base_cat("CPP-CVC"))
	    append_string(p, "c");
	  else if (daughters[k]->matching_base_cat("NP-PD") ||
		   daughters[k]->matching_base_cat("CNP-PD"))
	    append_string(p, "n");
	  else if (daughters[k]->matching_base_cat("S-OC") ||
		   daughters[k]->matching_base_cat("CS-OC"))
	    append_string(p, "s");
	}
	if (!is_new_subcat(daughters[i]->daughters[0]->label, 
			   daughters[i]->label, features))
	  daughters[i]->insert_feature(features);
	return;
      }
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::process                                                  */
/*                                                                 */
/*FE****************************************************************/

void Tree::process()

{
  static unsigned long N=0;
  if (Verbose)
    fprintf(stderr,"\r%lu", ++N);

  if (UnaryBranches) {
    if (print_unary_branches())
      print();
    return;
  }

  if (UndoMarkov)
    undo_markov();

  if (RestoreGF)
    restore_gf();

  if (ConvertTraces || EmptyCategories) {
    Trace_Number = 0;
    bool remaining_traces;
    do {
      remaining_traces = false;
    } while (convert(remaining_traces) && remaining_traces);

    if (remaining_traces)
      fprintf(stderr,"Warning: Some slash feature could not be translated to a trace index!\n");

    delete_slash_features();
    if (EmptyCategories)
      print_ec();
    else {
      if (DeleteChains)
	delete_chains();
      if (ReorderWords) {
	fill_gaps();
	reorder();
      }
      print();
    }
    return;
  }

  if (DeleteChains) {
    delete_chains();
    print();
    return;
  }

  if (UndoMarkov || RestoreGF) {
    print();
    return;
  }

  if (ReorderWords) {
    fill_gaps();
    reorder();
    print();
    return;
  }

  if (WithHeads)
    assign_heads();

  if (Governors) {
    print_govs();
    putchar('\n');
    return;
  }

  delete_features();
  convert_conjunctions();
  map_categories();
  correct_labels();

  insert_chain_rules();
  add_VP_features();
  add_nohead_feature();
  add_adverb_features();
  add_CARD_features();
  add_PP_features();
  add_NP_features();
  add_KOUS_features();
  add_punctuation_features();
  add_AP_features();
  add_KON_features();
  //add_subcat_feature();
  //add_domV_feature();
  add_nosubj_feature();

  if (PrintTraces) {
    add_slashes();
    collapse_traces();
  }
  delete_indices();

  delete_traces();

  if (PrintRules)
    add_rules();
  else
    print();
}
