
/*MA****************************************************************/
/*                                                                 */
/*     File: tree.h                                                */
/*   Author: Helmut Schmid                                         */
/*  Purpose:                                                       */
/*  Created: Tue Jan 21 13:27:46 2003                              */
/* Modified: Mon Mar 17 10:39:47 2014 (schmid)                     */
/*                                                                 */
/*ME****************************************************************/


#include <stdlib.h> 
#include <stdio.h>
#include <vector>
using std::vector;

#include "symbol-table.h"

extern bool Verbose;
extern bool PrintRules;
extern bool WithHeads;
extern bool PrettyPrint;
extern bool DeleteChains;
extern bool PrintTraces;
extern bool UndoMarkov;
extern bool UnaryBranches;
extern bool EmptyCategories;
extern bool ConvertTraces;
extern bool ReorderWords;
extern bool Governors;
extern bool RestoreGF;
extern int  Trace_Number;

extern char *MapFileName;
extern char *PrepFileName;

extern FILE *outfile, *lexfile;


class Tree;

typedef vector<Tree*> Trees;

void read_mapping( char*, SymbolMap& );

extern SymbolMap CatMap;
extern SymbolMap PrepMap;


/*****************  class Tree  ************************************/

class Tree {

private:
  void assign_heads();
  void print_subtree(int);
  int print_ec2(int);

  void change_label( const char* );
  void change_base_label( const char * );
  void change_gf_label( const char *l );
  bool is_trace();
  bool is_first_daughter();
  bool is_last_daughter();
  bool is_preposition();
  bool has_feature( const char *f );
  Tree *first_daughter() { return daughters[0]; };
  Tree *last_daughter() { return daughters[daughters.size()-1]; };
  Tree *predecessor();
  Tree *successor();
  void strip_features();
  const char *base_cat();
  bool matching_suffix( const char* );
  bool matching_base_cat( const char* );
  bool matching_category( const char* );
  bool has_trace_feature( char, const char* );
  char matching_trace( const char* );
  void append_slash( char*, char, bool );
  char append_slashes( char*, const char*, bool& );
  void insert_chain_node( const char *cat, size_t pos );

  bool matching_prefix( const char *prefix );
  void add_markup( const char *s, char d );
  void insert_feature( const char *s );
  void insert_uniq_feature( const char *s );

  void delete_features();
  void map_categories(); 
  void insert_chain_rules();
  bool add_rel_features();

  void add_slashes();
  void mark_as_trace();
  bool collapse_traces();
  bool delete_traces();
  void delete_indices();
  void add_rules();
  void print_govs();
  const char *cut_off_indices(); 
  void undo_markov();
  void restore_gf();
  void print_ec();
  bool print_unary_branches();

  char *slash_cat( char *c );
  char *find_trace( char*, char, bool& );
  bool delete_slashes_down( char *sc, char d );
  bool delete_slashes_up( char *sc, char d );
  bool convert( bool& );
  void delete_slash_features();

  void add_trace_feature( char c );
  void correct_labels();
  void convert_conjunctions();
  void add_adverb_features();
  void add_punctuation_features();
  void add_NP_features();
  void add_PP_features();
  void add_AP_features();
  void add_CARD_features();
  void add_KOUS_features();
  void add_VP_features();
  void add_KON_features();
  void add_nohead_feature();
  bool add_domV_feature();
  void add_subcat_feature();
  void add_nosubj_feature();
  void delete_chains();

  bool is_mod( bool );
  bool is_wh_phrase();
  void move_node( int from, int to );
  void reverse_order( int i, int k );
  void reorder();
  Tree *find_filler( char *index );
  bool fill_gap( Tree *node, char *index );
  void fill_gaps();

public:
  char  *label;
  Trees daughters;
  Tree  *mother;
  Tree  *head_daughter;
  Tree  *head;
  Tree  *head2;

  Tree( char *l ) { label = l; mother = head_daughter = head = head2 = NULL; };

  Tree( char *l, char *l2 ) {
    label = l;
    daughters.push_back(new Tree(l2));
    mother = head_daughter = head = head2 = NULL; 
  };

  Tree( char *l, Trees *d ) { 
    label = l;
    daughters = *d;
    delete d;
    for( size_t i=0; i<daughters.size(); i++ )
      daughters[i]->mother = this;
    mother = head_daughter = head = head2 = NULL; 
  };

  ~Tree() {
    free(label);
    for( size_t i=0; i<daughters.size(); i++ )
      delete daughters[i];
  }

  void process();
  void print();
};

void print_grammar();


char *last_char( char *s );
bool index_start( const char *s );
const char *last_char( const char *s );

void read_head_rules( char *filename );
bool strpref( const char *s, const char *p );
vector<char*> &split_name( char *s, size_t nmax );

void init();
char *find_index( char *label );
